unit Spellers;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  CSAPI, Registry, StdCtrls, ComCtrls, RichEdit, SpellRes, Langs;

type
  TSpellOption = (spoSuggestFromUserDict, spoIgnoreAllCaps, spoIgnoreMixedDigits,
                  spoIgnoreRomanNumerals, spoFindUncappedSentences,
                  spoFindMissingSpaces, spoFindRepeatWord, spoFindExtraSpaces,
                  spoFindSpacesBeforePunc, spoFindSpacesAfterPunc, spoRateSuggestions,
                  spoFindInitialNumerals);

  TSpellOptions = set of TSpellOption;

  TSpellCommand = (scVerifyWord, scVerifyBuffer, scSuggest, scSuggestMore,
                   scHyphInfo, scWildCard, scAnagram);

  TSpellReturnCode = (srNoErrors, srUnknownInputWord, srReturningChangeAlways,
                      srReturningChangeOnce, srInvalidHyphenation,
                      srErrorCapitalization, srWordConsideredAbbreviation,
                      srHyphChangesSpelling, srNoMoreSuggestions,
                      srMoreInfoThanBufferCouldHold, srNoSentenceStartCap,
                      srRepeatWord, srExtraSpaces, srMissingSpace,
                      srInitialNumeral);

  TSpellMode = (smInteractive, smBackground, smParallel);

  TMisspellFontOption = (mfoName, mfoSize, mfoHeight, mfoColor, mfoPitch,
                         mfoStyle);

  TMisspellFontOptions = set of TMisspellFontOption;

  TMisspellFont = class(TFont)
  private
    fOptions: TMisspellFontOptions;
  published
    property MisspellOptions: TMisspellFontOptions read fOptions write fOptions;
  end;

  TMisspellEvent = procedure (Sender: TObject; SRC: TSpellReturnCode;
                              BufPos, Len: Integer) of object;
  TChangeTextEvent = procedure(Sender: TObject; BufPos, Len: Integer;
                               NewWord: String) of object;

  TSpellerDialog = class;

  TSpellChecker = class(TComponent)
  private
    SpellInstance: THandle;
    DLLName: String;
    LexName: String;
    UserDict: String;
    UserExc: String;
    FCS: Integer;
    FStrings: TStrings;
    FOptions: TSpellOptions;
    FLID: TLanguage;
    FActive: Boolean;
    FFinishMessage: String;
    FOnMisspell: TMisspellEvent;
    FOnChangeText: TChangeTextEvent;
    FOnOpen: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FShowFinishMessage: Boolean;
    FSpellMode: TSpellMode;
    FCaption: TCaption;
    FMemo: TCustomMemo;
    FThreads: TList;
    FFont: TMisspellFont;
  protected
    SpecChars: TWSC;
    InputBuffer: TSIB;
    ResultBuffer: TSRB;
    Mdrs: TMDRS;
    Udr: TUDR;
    Handle: TSPLID;
    CheckStr: String;
    CheckTxt: TStrings;
  { Spell functions}
    SpellVer: TSpellVerFunc;
    SpellInit: TSpellInitFunc;
    SpellOptions: TSpellOptionsFunc;
    SpellCheck: TSpellCheckFunc;
    SpellTerminate: TSpellTerminateFunc;
    SpellVerifyMdr: TSpellVerifyMdrFunc;
    SpellOpenMdr: TSpellOpenMdrFunc;
    SpellOpenUdr: TSpellOpenUdrFunc;
    SpellAddUdr: TSpellAddUdrFunc;
    SpellAddChangeUdr: TSpellAddChangeUdrFunc;
    SpellDelUdr: TSpellDelUdrFunc;
    SpellClearUdr: TSpellClearUdrFunc;
    SpellGetSizeUdr: TSpellGetSizeUdrFunc;
    SpellGetListUdr: TSpellGetListUdrFunc;
    SpellCloseMdr: TSpellCloseMdrFunc;
    SpellCloseUdr: TSpellCloseUdrFunc;
    procedure SetOptions(Value: TSpellOptions);
    procedure SetLocale(Value: TLanguage);
    procedure SetActive(Value: Boolean);
    procedure CheckInteractive(Text: TStrings);
    procedure CheckAuto(Text: TStrings);
    procedure CheckParallel(Text: TStrings);
    function CheckBuf: TSpellReturnCode;
    procedure ChangeOnce(Word: String);
    procedure Change(Word: String);
    procedure ChangeAlways(Word: String);
    procedure Delete;
    procedure Add;
    procedure Ignore;
    procedure IgnoreAlways;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Open;
    procedure Close;
    procedure CheckText(Text: TStrings);
    procedure CheckMemo(Memo: TCustomMemo);
    function IsKnownWord(Word: String): Boolean;
    procedure GetVariants(Word: String; Variants: TStrings);
    property Active: Boolean read FActive write SetActive;
  published
    property Options: TSpellOptions read FOptions write SetOptions;
    property Language: TLanguage read FLID write SetLocale;
    property FinishMessage: String read FFinishMessage write FFinishMessage;
    property OnMisspelling: TMisspellEvent read FOnMisspell write FOnMisspell;
    property OnChangeText: TChangeTextEvent read FOnChangeText write FOnChangeText;
    property OnOpen: TNotifyEvent read FOnOpen write FOnOpen;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property ShowFinishMessage: Boolean read FShowFinishMessage write FShowFinishMessage;
    property SpellMode: TSpellMode read FSpellMode write FSpellMode default smInteractive;
    property Caption: TCaption read FCaption write FCaption;
    property MisspellFont: TMisspellFont read FFont write FFont;
  end;

  ESpellError = class(Exception);

  TSpellerDialog = class(TForm)
    InfoMsg: TLabel;
    Misspelling: TRichEdit;
    Label2: TLabel;
    Variants: TListBox;
    ChangeButton: TButton;
    ChangeAllButton: TButton;
    SkipButton: TButton;
    SkipAllButton: TButton;
    AddButton: TButton;
    CancelButton: TButton;
    CancelEdit: TButton;
    DelButton: TButton;
    procedure FormCreate(Sender: TObject);
    procedure DelButtonClick(Sender: TObject);
    procedure SkipButtonClick(Sender: TObject);
    procedure SkipAllButtonClick(Sender: TObject);
    procedure AddButtonClick(Sender: TObject);
    procedure ChangeButtonClick(Sender: TObject);
    procedure ChangeAllButtonClick(Sender: TObject);
    procedure MisspellingProtectChange(Sender: TObject; StartPos,
      EndPos: Integer; var AllowChange: Boolean);
    procedure MisspellingChange(Sender: TObject);
    procedure CancelEditClick(Sender: TObject);
  private
    { Private declarations }
    Checker: TSpellChecker;
    FText: TStrings;
    FShowing: Boolean;
    ErrorKind: TSpellReturnCode;
    procedure ContinueCheck;
    procedure ShowForChange(Msg: TCaption);
    procedure ShowForDelete;
    procedure ShowForEdit(Msg: TCaption);
    procedure ShowMisspelling;
    procedure GetHotArea(var SS, SL: Integer);
  public
    { Public declarations }
    constructor Create(SpellChecker: TSpellChecker; Text: TStrings);
  end;

var
  SpellerDialog: TSpellerDialog;

const
  langAfrikaans = TLanguage(1078);
  langAlbanian = TLanguage(1052);
  langArabic = TLanguage(1025);
  langBasque = TLanguage(1069);
  langBelgianDutch = TLanguage(2067);
  langBelgianFrench = TLanguage(2060);
  langBrazilianPortuguese = TLanguage(1046);
  langBulgarian = TLanguage(1026);
  langByelorussian = TLanguage(1059);
  langCatalan = TLanguage(1027);
  langCroatian = TLanguage(1050);
  langCzech = TLanguage(1029);
  langDanish = TLanguage(1030);
  langDutch = TLanguage(1043);
  langEnglishAUS = TLanguage(3081);
  langEnglishCanadian = TLanguage(4105);
  langEnglishNewZealand = TLanguage(5129);
  langEnglishSouthAfrica = TLanguage(7177);
  langEnglishUK = TLanguage(2057);
  langEnglishUS = TLanguage(1033);
  langEstonian = TLanguage(1061);
  langFaeroese = TLanguage(1080);
  langFarsi = TLanguage(1065);
  langFinnish = TLanguage(1035);
  langFinnishSwedish = TLanguage(2077);
  langFrench = TLanguage(1036);
  langFrenchCanadian = TLanguage(3084);
  langGerman = TLanguage(1031);
  langGreek = TLanguage(1032);
  langHebrew = TLanguage(1037);
  langHungarian = TLanguage(1038);
  langItalian = TLanguage(1040);
  langIcelandic = TLanguage(1039);
  langIndonesian = TLanguage(1057);
  langJapanese = TLanguage(1041);
  langKorean = TLanguage(1042);
  langLatvian = TLanguage(1062);
  langLithuanian = TLanguage(1063);
  langMacedonian = TLanguage(1071);
  langMalaysian = TLanguage(1086);
  langMexicanSpanish = TLanguage(2058);
  langNorwegianBokmol = TLanguage(1044);
  langNorwegianNynorsk = TLanguage(2068);
  langPolish = TLanguage(1045);
  langPortuguese = TLanguage(2070);
  langRomanian = TLanguage(1048);
  langRussian = TLanguage(1049);
  langSerbianCyrillic = TLanguage(3098);
  langSerbianLatin = TLanguage(2074);
  langSesotho = TLanguage(1072);
  langSimplifiedChinese = TLanguage(2052);
  langSlovak = TLanguage(1051);
  langSlovenian = TLanguage(1060);
  langSpanish = TLanguage(1034);
  langSpanishModernSort = TLanguage(3082);
  langSwedish = TLanguage(1053);
  langSwissFrench = TLanguage(4108);
  langSwissGerman = TLanguage(2055);
  langSwissItalian = TLanguage(2064);
  langThai = TLanguage(1054);
  langTraditionalChinese = TLanguage(1028);
  langTsonga = TLanguage(1073);
  langTswana = TLanguage(1074);
  langTurkish = TLanguage(1055);
  langUkrainian = TLanguage(1058);
  langVenda = TLanguage(1075);
  langVietnamese = TLanguage(1066);
  langXhosa = TLanguage(1076);
  langZulu = TLanguage(1077);

procedure Register;
function Paragraphs(Memo: TCustomMemo): TStrings;
function UnwrappedText(Memo: TCustomMemo): String;
function GetSpellLanguages(Languages: TStrings; Option: TLangOption): Integer;

implementation

{$R *.DFM}

type
  TSpellThread = class(TThread)
  private
    { Private declarations }
    FSpeller: TSpellChecker;
    FText: TStrings;
    SIB: TSIB;
    SRB: TSRB;
    Pos: Integer;
    FMemo: TCustomMemo;
    procedure CallEvent;
    procedure FirstCall;
    procedure NextCall;
    procedure Finish;
  protected
    constructor Create(Speller: TSpellChecker; Text: TStrings; Memo: TCustomMemo);
    destructor Destroy; override;
    procedure Execute; override;
  end;


procedure Register;
begin
  RegisterComponents('Additional', [TSpellChecker]);
end;

function Paragraphs(Memo: TCustomMemo): TStrings;
var
  Buf: PChar;
  Len: Integer;
begin
  Len:= GetWindowTextLength(Memo.Handle);
  Buf:= AllocMem(Len+1);
  GetWindowText(Memo.Handle, Buf, Len+1);
  Result:= TStringList.Create;
  Result.Text:= Buf;
  FreeMem(Buf);
end;

function UnwrappedText(Memo: TCustomMemo): String;
var
  Buf: PChar;
  Len: Integer;
begin
  Len:= GetWindowTextLength(Memo.Handle);
  Buf:= AllocMem(Len+1);
  GetWindowText(Memo.Handle, Buf, Len+1);
  Result:= Buf;
  FreeMem(Buf);
end;

procedure ApplyFont(Memo: TCustomRichEdit; Font: TMisspellFont);
begin
  if mfoName in Font.MisspellOptions then
    Memo.SelAttributes.Name:= Font.Name;
  if mfoSize in Font.MisspellOptions then
    Memo.SelAttributes.Size:= Font.Size;
  if mfoHeight in Font.MisspellOptions then
    Memo.SelAttributes.Height:= Font.Height;
  if mfoColor in Font.MisspellOptions then
    Memo.SelAttributes.Color:= Font.Color;
  if mfoPitch in Font.MisspellOptions then
    Memo.SelAttributes.Pitch:= Font.Pitch;
  if mfoStyle in Font.MisspellOptions then
    Memo.SelAttributes.Style:= Font.Style;
end;

constructor TSpellChecker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Language:= GetSystemDefaultLCID;
  FMemo:= nil;
  FActive:= False;
  FOptions:= [spoSuggestFromUserDict, spoIgnoreAllCaps, spoIgnoreMixedDigits,
              spoIgnoreRomanNumerals];
  FCS:= -1;
  FStrings:= nil;
  FShowFinishMessage:= True;
  FThreads:= TList.Create;
  FFinishMessage:= spsFinish;
  FCaption:= spsCaption;
  FFont:= TMisspellFont.Create;
  FFont.fOptions:= [mfoColor];
  FFont.Color:= clRed;
  with SpecChars do
    begin
      bIgnore:= #0;
      bHyphenHard:= #45;
      bHyphenSoft:= #31;
      bHyphenNonBreaking:= #30;
      bEmDash:= #151;
      bEnDash:= #150;
      bEllipsis:= #133;
      rgLineBreak:= #11#10;
      rgParaBreak:= #13#10;
    end;
end;

destructor TSpellChecker.Destroy;
begin
  if FActive then
    Close;
  FFont.Free;
  FThreads.Free;
  inherited;
end;

procedure CheckSR(SR: TSEC);
begin
  if SR<>secNoErrors then
    raise ESpellError.CreateFmt(spsError, [SR]);
end;

procedure TSpellChecker.Open;
var
  Registry: TRegistry;
  UdrRO: Boolean;
  Lid: TLID;
  NotFound: Boolean;
begin
  if FActive then
    Exit;
  if csDesigning in ComponentState then
    Exit;
  Registry:= TRegistry.Create;
  Registry.RootKey:= HKEY_LOCAL_MACHINE;
  NotFound:= False;
  if Registry.OpenKey(
     Format('\SOFTWARE\Microsoft\Shared Tools\Proofing Tools\Spelling\%d\Normal', [FLID]),
     False) then
    begin
      DLLName:= Registry.ReadString('Engine');
      LexName:= Registry.ReadString('Dictionary');
    end
  else
    NotFound:= True;
  if Registry.OpenKey(
     '\SOFTWARE\Microsoft\Shared Tools\Proofing Tools\Custom Dictionaries',
     False) then
    begin
      UserDict:= Registry.ReadString('1');
      if UserDict='' then
        begin
          UserDict:= ExtractFilePath(LexName)+IntToStr(FLID)+'.dic';
          Registry.WriteString('1', UserDict);
        end;
    end;
  Registry.Free;
  if NotFound then
    Exit;
  Factive:= True;
  try
    SpellInstance:= LoadLibrary(PChar(DllName));
  except
    raise ESpellError.CreateFmt(spsErrorLoad, [DllName]);
  end;
  try
    @SpellVer:= GetProcAddress(SpellInstance, 'SpellVer');
    @SpellInit:= GetProcAddress(SpellInstance, 'SpellInit');
    @SpellOptions:= GetProcAddress(SpellInstance, 'SpellOptions');
    @SpellCheck:= GetProcAddress(SpellInstance, 'SpellCheck');
    @SpellTerminate:= GetProcAddress(SpellInstance, 'SpellTerminate');
    @SpellVerifyMdr:= GetProcAddress(SpellInstance, 'SpellVerifyMdr');
    @SpellOpenMdr:= GetProcAddress(SpellInstance, 'SpellOpenMdr');
    @SpellOpenUdr:= GetProcAddress(SpellInstance, 'SpellOpenUdr');
    @SpellAddUdr:= GetProcAddress(SpellInstance, 'SpellAddUdr');
    @SpellAddChangeUdr:= GetProcAddress(SpellInstance, 'SpellAddChangeUdr');
    @SpellDelUdr:= GetProcAddress(SpellInstance, 'SpellDelUdr');
    @SpellClearUdr:= GetProcAddress(SpellInstance, 'SpellClearUdr');
    @SpellGetSizeUdr:= GetProcAddress(SpellInstance, 'SpellGetSizeUdr');
    @SpellGetListUdr:= GetProcAddress(SpellInstance, 'SpellGetListUdr');
    @SpellCloseMdr:= GetProcAddress(SpellInstance, 'SpellCloseMdr');
    @SpellCloseUdr:= GetProcAddress(SpellInstance, 'SpellCloseUdr');
  except
    FreeLibrary(SpellInstance);
    raise ESpellError.CreateFmt(spsErrorLoad, [DLLName]);
  end;
  CheckSR(SpellInit(Handle, SpecChars));
  CheckSR(SpellOptions(Handle, Word(FOptions)));
  if Assigned(FOnOpen) then
    FOnOpen(Self);
  CheckSR(SpellVerifyMdr(PChar(LexName), FLID, LID));
  CheckSR(SpellOpenMdr(Handle, PChar(LexName), nil, False, True, FLID, Mdrs));
  CheckSR(SpellOpenUdr(Handle, PChar(UserDict), True, IgnoreAlwaysProp, Udr, UdrRO));
end;

procedure TSpellChecker.Close;
var
  I: Integer;
begin
  if not FActive then
    Exit;
  FActive:= False;
  if csDesigning in ComponentState then
    Exit;
  for I:= FThreads.Count-1 downto 0 do
    TSpellThread(FThreads.Items[I]).Terminate;
  CheckSR(SpellTerminate(Handle, True));
  try
    FreeLibrary(SpellInstance);
  except
    raise ESpellError.CreateFmt(spsErrorUnload, [DLLName]);
  end;
  if Assigned(FOnClose) then
    FOnClose(Self);
end;

procedure TSpellChecker.SetOptions(Value: TSpellOptions);
begin
  FOptions:= Value;
  if not Active then
    Exit;
  if csDesigning in ComponentState then
    Exit;
  CheckSR(SpellOptions(Handle, Integer(Word(FOptions))));
end;

procedure TSpellChecker.SetLocale(Value: TLanguage);
begin
  if Active or not IsValidLocale(Value, LCID_Installed) then
    Exit;
  FLID:= Value;
end;

procedure TSpellChecker.SetActive(Value: Boolean);
begin
  if Value xor FActive then
    if Value then
      Open
    else
      Close;
end;

procedure TSpellChecker.CheckInteractive(Text: TStrings);
var
  I: Integer;
begin
  CheckTxt:= Text;
  CheckStr:= '';
  for I:= 0 to Text.Count-1 do
    CheckStr:= CheckStr+Text[I]+#13#10;
  with InputBuffer do
    begin
      cch:= Length(CheckStr);
      cMdr:= 1;
      cUdr:= 1;
      wSpellState:= fssNoStateInfo;
      lrgch:= @CheckStr[1];
      lrgMdr:= @Mdrs.MDR;
      lrgUdr:= @Udr;
    end;
  with ResultBuffer do
    begin
      cch:= 1024;
      lrgsz:= AllocMem(1024);
      lrgbRating:= AllocMem(255);
      cbRate:= 255;
    end;
  with TSpellerDialog.Create(Self, Text) do
    begin
      ContinueCheck;
      Free;
    end;
  FreeMem(ResultBuffer.lrgsz);
  FreeMem(ResultBuffer.lrgbRating);
  if ShowFinishMessage and (FinishMessage<>'') then
    MessageBoxEx(Application.Handle, PChar(FinishMessage), PChar(spsFinishCaption),
               mb_Ok+mb_IconInformation, FLID);
end;

procedure TSpellChecker.CheckAuto(Text: TStrings);
var
  SIB: TSIB;
  SRB: TSRB;
  S: String;
  I, Pos: Integer;
begin
  S:= '';
  for I:= 0 to Text.Count-1 do
    S:= S+Text[I]+#13#10;
  with SIB do
    begin
      cch:= Length(S);
      cMdr:= 1;
      cUdr:= 1;
      wSpellState:= fssStartsSentence;
      lrgch:= @S[1];
      lrgMdr:= @Mdrs.MDR;
      lrgUdr:= @Udr;
    end;
  SRB.cch:= 255;
  SRB.lrgsz:= AllocMem(255);
  CheckSR(SpellCheck(Handle, sccVerifyBuffer, SIB, SRB));
  Pos:= 0;
  while SRB.scrs<>scrsNoErrors do
    begin
      Inc(Pos, SRB.ichError);
      if Assigned(FMemo) and (FMemo is TCustomRichEdit) and Assigned(FFont) then
        begin
          FMemo.SelStart:= Pos;
          FMemo.SelLength:= SRB.cchError;
          ApplyFont(TCustomRichEdit(FMemo), FFont);
        end;
      if Assigned(FOnMisspell) then
        FOnMisspell(Self, TSpellReturnCode(SRB.scrs), Pos, SRB.cchError);
      Inc(Pos, SRB.cchError);
      SIB.wSpellState:= fssIsContinued;
      Inc(SIB.lrgch, SRB.ichError+SRB.cchError);
      Dec(SIB.cch, SRB.ichError+SRB.cchError);
      CheckSR(SpellCheck(Handle, sccVerifyBuffer, SIB, SRB));
    end;
  FreeMem(SRB.lrgsz);
end;

procedure TSpellChecker.CheckParallel(Text: TStrings);
begin
  FThreads.Add(TSpellThread.Create(Self, Text, FMemo));
end;

procedure TSpellChecker.CheckText(Text: TStrings);
begin
  if not Active then
    Exit;
  case SpellMode of
  smInteractive:
    CheckInteractive(Text);
  smBackground:
    CheckAuto(Text);
  smParallel:
    CheckParallel(Text);
  end;
end;

function TSpellChecker.IsKnownWord(Word: String): Boolean;
var
  SIB: TSIB;
  SRB: TSRB;
begin
  Result:= True;
  if not Active then
    Exit;
  with SIB do
    begin
      cch:= Length(Word);
      cMdr:= 1;
      cUdr:= 1;
      wSpellState:= fssNoStateInfo;
      lrgch:= @Word[1];
      lrgMdr:= @Mdrs.MDR;
      lrgUdr:= @Udr;
    end;
  CheckSR(SpellCheck(Handle, sccVerifyWord, SIB, SRB));
  Result:= SRB.scrs=scrsNoErrors;
end;

procedure TSpellChecker.GetVariants(Word: String; Variants: TStrings);
var
  SIB: TSIB;
  SRB: TSRB;
  Buf: array[0..2047]of Char;
  Ratings: array[0..255]of Byte;
  P: PChar;
begin
  if not Active then
    Exit;
  Variants.Clear;
  with SIB do
    begin
      cch:= Length(Word);
      cMdr:= 1;
      cUdr:= 1;
      wSpellState:= fssNoStateInfo;
      lrgch:= @Word[1];
      lrgMdr:= @Mdrs.MDR;
      lrgUdr:= @Udr;
    end;
  with SRB do
    begin
      cch:= 1024;
      lrgsz:= @Buf;
      lrgbRating:= @Ratings;
      cbRate:= 255;
    end;
  CheckSR(SpellCheck(Handle, sccSuggest, SIB, SRB));
  while SRB.scrs<>scrsNoMoreSuggestions do
    begin
      P:= SRB.lrgsz;
      while P^<>#0 do
        begin
          if Variants.IndexOf(P)=-1 then
            Variants.Add(P);
          while P^<>#0 do
            Inc(P);
          Inc(P);
        end;
      CheckSR(SpellCheck(Handle, sccSuggestMore, SIB, SRB));
    end;
end;

function TSpellChecker.CheckBuf: TSpellReturnCode;
begin
  CheckSR(SpellCheck(handle, sccVerifyBuffer, InputBuffer, ResultBuffer));
  Result:= TSpellReturnCode(ResultBuffer.scrs);
  if Assigned(FMemo) and (ResultBuffer.scrs<>scrsNoErrors) then
    begin
      FMemo.SelStart:= InputBuffer.lrgch-@CheckStr[1]+ResultBuffer.ichError;
      FMemo.SelLength:= ResultBuffer.cchError;
    end;
end;

procedure TSpellChecker.ChangeOnce(Word: String);
var
  P: Integer;
begin
  CheckSR(SpellAddChangeUdr(Handle, udrChangeOnce,
      PChar(Copy(InputBuffer.lrgch, ResultBuffer.ichError+1, ResultBuffer.cchError)),
      PChar(Word)));
  P:= InputBuffer.lrgch-@CheckStr[1];
  CheckStr:= Copy(CheckStr, 1, P+ResultBuffer.ichError)+
             Word+
             Copy(CheckStr, P+ResultBuffer.ichError+ResultBuffer.cchError+1, Length(CheckStr));
  if Assigned(FMemo) then
    FMemo.SelText:= Word;
  if Assigned(FOnChangeText) then
    FOnChangeText(Self, P+ResultBuffer.ichError, ResultBuffer.cchError, Word);
  InputBuffer.lrgch:= @CheckStr[1+P+ResultBuffer.ichError];
  InputBuffer.cch:= StrLen(InputBuffer.lrgch);
end;

procedure TSpellChecker.Change(Word: String);
var
  P: Integer;
begin
  P:= InputBuffer.lrgch-@CheckStr[1];
  CheckStr:= Copy(CheckStr, 1, P+ResultBuffer.ichError)+
             Word+
             Copy(CheckStr, P+ResultBuffer.ichError+ResultBuffer.cchError+1, Length(CheckStr));
  if Assigned(FMemo) then
    FMemo.SelText:= Word;
  if Assigned(FOnChangeText) then
    FOnChangeText(Self, P+ResultBuffer.ichError, ResultBuffer.cchError, Word);
  InputBuffer.lrgch:= @CheckStr[1+P+ResultBuffer.ichError];
  InputBuffer.cch:= StrLen(InputBuffer.lrgch);
end;

procedure TSpellChecker.ChangeAlways(Word: String);
var
  P: Integer;
begin
  CheckSR(SpellAddChangeUdr(Handle, udrChangeAlways,
      PChar(Copy(InputBuffer.lrgch, ResultBuffer.ichError+1, ResultBuffer.cchError)),
      PChar(Word)));
  P:= InputBuffer.lrgch-@CheckStr[1];
  CheckStr:= Copy(CheckStr, 1, P+ResultBuffer.ichError)+
             Word+
             Copy(CheckStr, P+ResultBuffer.ichError+ResultBuffer.cchError+1, Length(CheckStr));
  if Assigned(FMemo) then
    FMemo.SelText:= Word;
  if Assigned(FOnChangeText) then
    FOnChangeText(Self, P+ResultBuffer.ichError, ResultBuffer.cchError, Word);
  InputBuffer.lrgch:= @CheckStr[1+P+ResultBuffer.ichError+Length(Word)];
  InputBuffer.cch:= StrLen(InputBuffer.lrgch);
end;

procedure TSpellChecker.Delete;
var
  P: Integer;
begin
  P:= InputBuffer.lrgch-@CheckStr[1];
  System.Delete(CheckStr, P+ResultBuffer.ichError, ResultBuffer.cchError+1);
  if Assigned(FMemo) then
    begin
      FMemo.SelStart:= P+ResultBuffer.ichError-1;
      FMemo.SelLength:= ResultBuffer.cchError+1;
      FMemo.SelText:= '';
    end;
  if Assigned(FOnChangeText) then
    FOnChangeText(Self, P+ResultBuffer.ichError-1, ResultBuffer.cchError+1, '');
  InputBuffer.lrgch:= @CheckStr[1+P+ResultBuffer.ichError];
  InputBuffer.cch:= StrLen(InputBuffer.lrgch);
end;

procedure TSpellChecker.Add;
begin
  CheckSR(SpellAddUdr(Handle, Udr,
      PChar(Copy(InputBuffer.lrgch, ResultBuffer.ichError+1, ResultBuffer.cchError))));
  Inc(InputBuffer.lrgch, ResultBuffer.ichError+ResultBuffer.cchError);
  Dec(InputBuffer.cch, ResultBuffer.ichError+ResultBuffer.cchError);
end;

procedure TSpellChecker.Ignore;
begin
  Inc(InputBuffer.lrgch, ResultBuffer.ichError+ResultBuffer.cchError);
  Dec(InputBuffer.cch, ResultBuffer.ichError+ResultBuffer.cchError);
end;

procedure TSpellChecker.IgnoreAlways;
begin
  CheckSR(SpellAddUdr(Handle, udrIgnoreAlways,
      PChar(Copy(InputBuffer.lrgch, ResultBuffer.ichError+1, ResultBuffer.cchError))));
  Inc(InputBuffer.lrgch, ResultBuffer.ichError+ResultBuffer.cchError);
  Dec(InputBuffer.cch, ResultBuffer.ichError+ResultBuffer.cchError);
end;

procedure TSpellChecker.CheckMemo(Memo: TCustomMemo);
var
  Para: TStrings;
begin
  Para:= Paragraphs(Memo);
  FMemo:= Memo;
  CheckText(Para);
  Para.Free;
  FMemo:= nil;
end;

constructor TSpellThread.Create(Speller: TSpellChecker; Text: TStrings;
                                Memo: TCustomMemo);
begin
  FSpeller:= Speller;
  FText:= TStringList.Create;
  FText.Text:= Text.Text;
  FMemo:= Memo;
  FreeOnTerminate:= True;
  inherited Create(False);
end;

destructor TSpellThread.Destroy;
begin
  if Assigned(FText) then
    FText.Free;
  inherited;
end;

procedure TSpellThread.CallEvent;
begin
  if Assigned(FMemo) and (FMemo is TCustomRichEdit) and Assigned(FSpeller.FFont) then
    begin
      FMemo.SelStart:= Pos;
      FMemo.SelLength:= SRB.cchError;
      ApplyFont(TCustomRichEdit(FMemo), FSpeller.FFont);
    end;
  with FSpeller do
    if Assigned(FOnMisspell) then
      FOnMisspell(FSpeller, TSpellReturnCode(SRB.scrs), Pos, SRB.cchError);
end;

procedure TSpellThread.FirstCall;
begin
  CheckSR(FSpeller.SpellCheck(FSpeller.Handle, sccVerifyBuffer, SIB, SRB));
end;

procedure TSpellThread.NextCall;
begin
  CheckSR(FSpeller.SpellCheck(FSpeller.Handle, sccVerifyBuffer, SIB, SRB));
end;

procedure TSpellThread.Finish;
begin
  FSpeller.FThreads.Delete(FSpeller.FThreads.IndexOf(Self));
end;

procedure TSpellThread.Execute;
var
  S: String;
  I: Integer;
begin
  S:= FText.Text;
  if Terminated then
    Exit;
  with SIB do
    begin
      cch:= Length(S);
      cMdr:= 1;
      cUdr:= 1;
      wSpellState:= fssStartsSentence;
      lrgch:= @S[1];
      lrgMdr:= @FSpeller.Mdrs.MDR;
      lrgUdr:= @FSpeller.Udr;
    end;
  SRB.cch:= 255;
  SRB.lrgsz:= AllocMem(255);
  if Terminated then
    begin
      Synchronize(Finish);
      Exit;
    end;
  Synchronize(FirstCall);
  if Terminated then
    begin
      Synchronize(Finish);
      Exit;
    end;
  Pos:= 0;
  while SRB.scrs<>scrsNoErrors do
    begin
      Inc(Pos, SRB.ichError);
      Synchronize(CallEvent);
      Inc(Pos, SRB.cchError);
      if Terminated then
        Break;
      SIB.wSpellState:= fssIsContinued;
      Inc(SIB.lrgch, SRB.ichError+SRB.cchError);
      Dec(SIB.cch, SRB.ichError+SRB.cchError);
      Synchronize(NextCall);
      if Terminated then
        Break;
    end;
  FreeMem(SRB.lrgsz);
  Synchronize(Finish);
end;

constructor TSpellerDialog.Create(SpellChecker: TSpellChecker; Text: TStrings);
begin
  Checker:= SpellChecker;
  FText:= Text;
  inherited Create(Checker.Owner);
end;

procedure TSpellerDialog.FormCreate(Sender: TObject);
begin
  Caption:= Format(Checker.Caption, [LanguageName(Checker.Language)]);
  Label2.Caption:= spsVariants;
  ChangeButton.Caption:= spsChange;
  ChangeAllButton.Caption:= spsChangeAll;
  SkipButton.Caption:= spsSkip;
  SkipAllButton.Caption:= spsSkipAll;
  DelButton.Caption:= spsDelete;
  CancelButton.Caption:= spsCancel;
  CancelEdit.Caption:= spsCancelEdit;
  Variants.Font.Charset:= CharSetFromLocale(Checker.Language);
end;

procedure TSpellerDialog.ContinueCheck;
begin
  ErrorKind:= Checker.CheckBuf;
  case Integer(ErrorKind) of
  scrsNoErrors:
    Close;
  scrsUnknownInputWord:
    ShowForChange(spsNotFound);
  scrsReturningChangeAlways:
    Checker.Change(Checker.ResultBuffer.lrgsz);
  scrsReturningChangeOnce:
    begin
      ShowForChange(spsNotFound);
      if Variants.Items.IndexOf(Checker.ResultBuffer.lrgsz)=-1 then
        Variants.Items.Insert(0, Checker.ResultBuffer.lrgsz);
    end;
  scrsInvalidHyphenation:
    ShowForChange(spsHyphen);
  scrsErrorCapitalization:
    ShowForChange(spsCaps);
  scrsWordConsideredAbbreviation:
    ShowForChange(spsAbbrev);
  scrsNoSentenceStartCap:
    ShowForChange(spsNoSentenceCap);
  scrsRepeatWord:
    ShowForDelete;
  scrsExtraSpaces:
    ShowForChange(spsExtraSpaces);
  scrsMissingSpace:
    ShowForEdit(spsMissingSpace);
  scrsInitialNumeral:
    ShowForEdit(spsInitialNumeral);
  end;
end;

procedure TSpellerDialog.DelButtonClick(Sender: TObject);
begin
  Checker.Delete;
  ContinueCheck;
end;

procedure TSpellerDialog.SkipButtonClick(Sender: TObject);
begin
  Checker.Ignore;
  ContinueCheck;
end;

procedure TSpellerDialog.SkipAllButtonClick(Sender: TObject);
begin
  Checker.IgnoreAlways;
  ContinueCheck;
end;

procedure TSpellerDialog.AddButtonClick(Sender: TObject);
begin
  Checker.Add;
  ContinueCheck;
end;

procedure TSpellerDialog.ChangeButtonClick(Sender: TObject);
var
  SS, SL: Integer;
begin
  if CancelEdit.Visible then
    begin
      GetHotArea(SS, SL);
      with Misspelling do
        begin
          SelStart:= SS;
          SelLength:= SL;
          Checker.Change(SelText);
        end;
      ContinueCheck;
    end
  else
    begin
      if Integer(ErrorKind)=scrsUnknownInputWord then
        Checker.ChangeOnce(Variants.Items[Variants.ItemIndex])
      else
         Checker.Change(Variants.Items[Variants.ItemIndex]);
     ContinueCheck;
    end;
end;

procedure TSpellerDialog.ChangeAllButtonClick(Sender: TObject);
begin
  Checker.ChangeAlways(Variants.Items[Variants.ItemIndex]);
  ContinueCheck;
end;

function DelExtraSpaces(const S: String): String;
var
  I: Integer;
begin
  I:= Length(S)-2;
  while S[I]=' ' do
    Dec(I);
  Result:= Copy(S, 1, I)+Copy(S, Length(S)-1, 2);
end;

function SentenceCapitalize(const S: String; Locale: TLanguage): String;
begin
  Result:= S;
  LCMapString(Locale, LCMAP_Uppercase, @Result[1], 1, @Result[1], 1);
end;

procedure TSpellerDialog.ShowForChange(Msg: TCaption);
var
  UnkWord: String;
begin
  ShowMisspelling;
  InfoMsg.Caption:= Msg;
  CancelEdit.Hide;
  DelButton.Hide;
  ChangeButton.Show;
  SkipButton.Enabled:= True;
  SkipAllButton.Enabled:= Integer(ErrorKind)=scrsUnknownInputWord;
  AddButton.Enabled:= Integer(ErrorKind)=scrsUnknownInputWord;
  UnkWord:= Copy(Checker.InputBuffer.lrgch, Checker.ResultBuffer.ichError+1,
                 Checker.ResultBuffer.cchError);
  case Integer(ErrorKind) of
  scrsUnknownInputWord,
  scrsReturningChangeOnce,
  scrsInvalidHyphenation,
  scrsErrorCapitalization:
    Checker.GetVariants(UnkWord, Variants.Items);
  scrsNoSentenceStartCap:
    begin
      Variants.Items.Clear;
      Variants.Items.Add(SentenceCapitalize(UnkWord, Checker.Language));
    end;
  scrsExtraSpaces:
    begin
      Variants.Items.Clear;
      Variants.Items.Add(DelExtraSpaces(UnkWord));
    end;
  end;
  Variants.ItemIndex:= 0;
  ChangeButton.Enabled:= Variants.ItemIndex=0;
  ChangeAllButton.Enabled:= (Variants.ItemIndex=0) and (Integer(ErrorKind)=scrsUnknownInputWord);
  Variants.Enabled:= Variants.ItemIndex=0;
  if not Visible then
    ShowModal;
end;

procedure TSpellerDialog.ShowForEdit(Msg: TCaption);
var
  UnkWord: String;
begin
  ShowMisspelling;
  InfoMsg.Caption:= Msg;
  CancelEdit.Hide;
  DelButton.Hide;
  ChangeButton.Show;
  SkipButton.Enabled:= True;
  SkipAllButton.Enabled:= false;
  AddButton.Enabled:= False;
  UnkWord:= Copy(Checker.InputBuffer.lrgch, Checker.ResultBuffer.ichError+1,
                 Checker.ResultBuffer.cchError);
  Variants.Items.Clear;
  ChangeButton.Enabled:= False;
  ChangeAllButton.Enabled:= False;
  Variants.Enabled:= False;
  if not Visible then
    ShowModal;
end;

procedure TSpellerDialog.ShowForDelete;
begin
  ShowMisspelling;
  InfoMsg.Caption:= spsRepeatedWord;
  CancelEdit.Hide;
  DelButton.Show;
  ChangeButton.Hide;
  ChangeAllButton.Enabled:= False;
  SkipButton.Show;
  SkipAllButton.Enabled:= False;
  AddButton.Enabled:= False;
  if not Visible then
    ShowModal;
end;

procedure TSpellerDialog.ShowMisspelling;
var
  CurStr: String;
  I: Integer;
  P, PS, PE: PChar;
begin
  FShowing:= True;
  P:= Checker.InputBuffer.lrgch+Checker.ResultBuffer.ichError;
  PS:= P;
  PE:= P;
  while (PS>@Checker.CheckStr[1]) and not (PS^ in [#13, #10]) do
    Dec(PS);
  if PS^ in [#13, #10] then
    Inc(PS);
  while not (PE^ in [#0, #13, #10]) do
    Inc(PE);
  with Misspelling do
    begin
      Lines.Clear;
      SelAttributes.Charset:= Variants.Font.CharSet;
      Lines.Add(Copy(PS, 1, PE-PS));
      SelectAll;
      SelAttributes.Color:= clWindowText;
      SelAttributes.Protected:= True;
      SelStart:= P-PS;
      SelLength:= Checker.ResultBuffer.cchError;
      SelAttributes.Protected:= False;
      SelAttributes.Color:= clRed;
      SelLength:= 0;
    end;
  FShowing:= False;
end;

procedure TSpellerDialog.MisspellingProtectChange(Sender: TObject;
  StartPos, EndPos: Integer; var AllowChange: Boolean);
begin
  AllowChange:= FShowing;
end;

procedure TSpellerDialog.MisspellingChange(Sender: TObject);
begin
  if FShowing then
    Exit;
  CancelEdit.Show;
  ChangeButton.Show;
  ChangeButton.Enabled:= True;
  ChangeAllButton.Enabled:= False;
  SkipAllButton.Enabled:= False;
  AddButton.Enabled:= False;
  DelButton.Hide;
  Variants.Enabled:= False;
end;

procedure TSpellerDialog.GetHotArea(var SS, SL: Integer);
var
  I: Integer;
begin
  with Misspelling do
    begin
      SelLength:= 0;
      for I:= 0 to Length(UnwrappedText(Misspelling))-1 do
        begin
          SelStart:= I;
          if not SelAttributes.Protected then
            begin
              SS:= I;
              Break;
            end;
        end;
      for I:= SS to Length(UnwrappedText(Misspelling)) do
        begin
          SelStart:= I;
          if SelAttributes.Protected then
            begin
              SL:= I-SS-1;
              Exit;
            end;
        end;
      SL:= Length(Misspelling.Lines[0])-SS;
    end;
end;

procedure TSpellerDialog.CancelEditClick(Sender: TObject);
var
  SS, SL: Integer;
  UnkWord: String;
begin
  GetHotArea(SS, SL);
  UnkWord:= Copy(Checker.InputBuffer.lrgch, Checker.ResultBuffer.ichError+1,
                 Checker.ResultBuffer.cchError);
  with Misspelling do
    begin
      SelStart:= SS;
      SelLength:= SL;
      SelText:= UnkWord;
      SelLength:= 0;
    end;
  case Integer(ErrorKind) of
  scrsUnknownInputWord,
  scrsReturningChangeOnce,
  scrsInvalidHyphenation,
  scrsErrorCapitalization,
  scrsNoSentenceStartCap,
  scrsExtraSpaces:
    begin
      DelButton.Hide;
      ChangeButton.Show;
      SkipButton.Enabled:= True;
      SkipAllButton.Enabled:= Integer(ErrorKind)=scrsUnknownInputWord;
      AddButton.Enabled:= Integer(ErrorKind)=scrsUnknownInputWord;
      ChangeAllButton.Enabled:= (Variants.ItemIndex=0) and (Integer(ErrorKind)=scrsUnknownInputWord);
      ChangeAllButton.Enabled:= Variants.ItemIndex=0;
      Variants.Enabled:= Variants.ItemIndex=0;
    end;
  scrsRepeatWord:
    begin
      DelButton.Show;
      ChangeButton.Hide;
      ChangeAllButton.Enabled:= False;
      SkipButton.Show;
      SkipAllButton.Enabled:= False;
      AddButton.Enabled:= False;
    end;
  scrsWordConsideredAbbreviation,
  scrsMissingSpace,
  scrsInitialNumeral:
    begin
      DelButton.Hide;
      ChangeButton.Show;
      SkipButton.Enabled:= True;
      SkipAllButton.Enabled:= false;
      AddButton.Enabled:= False;
      ChangeButton.Enabled:= False;
      ChangeAllButton.Enabled:= False;
      Variants.Enabled:= False;
    end;
  end;
  CancelEdit.Hide;
end;

function GetSpellLanguages(Languages: TStrings; Option: TLangOption): Integer;
var
  Registry: TRegistry;
  N, I: Integer;
  LCType: Integer;
  Buf: array[0..255]of Char;
begin
  Registry:= TRegistry.Create;
  Registry.RootKey:= HKEY_LOCAL_MACHINE;
  Registry.OpenKey('\SOFTWARE\Microsoft\Shared Tools\Proofing Tools\Spelling', False);
  Registry.GetKeyNames(Languages);
  Registry.Free;
  Result:= Languages.Count;
  case Option of
  loLocalized:
    LCType:= LOCALE_SLanguage;
  loEnglish:
    LCType:= LOCALE_SEngLanguage;
  loNative:
    LCType:= LOCALE_SNativeLangName;
  loAbbrev:
    LCType:= LOCALE_SAbbrevLangName;
  end;
  for I:= Languages.Count-1 downto 0 do
    begin
      N:= StrToIntDef(Languages[I], -1);
      if N=-1 then
        Languages.Delete(I)
      else
        begin
          Languages.Objects[I]:= Pointer(N);
          GetLocaleInfo(N, LCType, Buf, 255);
          Languages[I]:= Buf;
        end;
    end;
end;

end.
